/*
 * X11 GUI for slide puzzle game
 * Copyright © 2022 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <locale.h>
#include <getopt.h>

#include <Xm/XmAll.h>

#include "help.h"
#include "motif.h"
#include "motifopt.h"
#include "game.h"
#include "version.h"

#define TIMER_UPDATE_MS 33

#define PROGNAME "rrace"
static const char *progname = PROGNAME;
static const struct option lopts[] = { LOPTS_INITIALIZER, {0} };

static char * const default_resources[] = {
	"*title: RRace",
	"*game.height: 371", // 365 + 2*shadowThickness
	"*game.width: 498",

	"*game.XmFrame.shadowThickness: 3",
	"*game.XmFrame.shadowType: shadow_in",
	"*goalArea.leftOffset: 1",

	"*gameNew.accelerator: Ctrl<Key>N",
	"*gameNew.acceleratorText: Ctrl+N",
	"*gameExit.accelerator: Ctrl<Key>Q",
	"*gameExit.acceleratorText: Ctrl+Q",

	"*aboutDialog*pixmapTextPadding: 10",

	NULL
};

static void print_version(void)
{
	version_print_head("rrace-motif", stdout);
	puts("License GPLv3+: GNU GPL version 3 or any later version");
	puts("This is free software: you are free to change and redistribute it.");
	puts("There is NO WARRANTY, to the extent permitted by law.");
}

static void print_usage(FILE *f)
{
	fprintf(f, "Usage: %s [options]\n", progname);
	if (f != stdout)
		fprintf(f, "Try %s --help for more information.\n", progname);
}

static void print_help(void)
{
	struct lopt_help help = {0};
	const struct option *opt;

	print_usage(stdout);

	putchar('\n');
	puts("Options:");
	for (opt = lopts; opt->name; opt++) {
		if (!lopt_get_help(opt, &help))
			continue;
		help_print_option(opt, help.arg, help.desc, 20);
	}
	putchar('\n');

	printf("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
}

static Widget early_setup(XtAppContext *app, int argc, char **argv)
{
	Widget shell;
	int opt;

	/* Check for --help/--version early (before X connection) */
	opterr = 0;
	while ((opt = getopt_long_only(argc, argv, "", lopts, NULL)) != -1) {
		switch (opt) {
		case LOPT_VERSION:
			print_version();
			exit(EXIT_SUCCESS);
		case LOPT_HELP:
			print_help();
			exit(EXIT_SUCCESS);
		}
	}

	shell = XtOpenApplication(app, PACKAGE_TARNAME, NULL, 0,
	                          &argc, argv, (String *)default_resources,
	                          sessionShellWidgetClass, NULL, 0);

	opterr = optind = 1;
	while ((opt = getopt_long_only(argc, argv, "", lopts, NULL)) != -1) {
		switch (opt) {
		default:
			print_usage(stderr);
			exit(EXIT_FAILURE);
		}
	}

	if (argv[optind]) {
		fprintf(stderr, "%s: excess command-line arguments.\n",
		                progname);
		print_usage(stderr);
		exit(EXIT_FAILURE);
	}

	return shell;
}

static void timer_tick(void *data, XtIntervalId *id)
{
	struct app_state *state = data;
	XtAppContext app;

	if (state->board.x > 4) {
		/* Game is over */
		state->timer_tick = 0;
		return;
	}

	app = XtWidgetToApplicationContext(state->timer);
	ui_timer_update(state, game_elapsed(&state->board));
	state->timer_tick = XtAppAddTimeOut(app, TIMER_UPDATE_MS,
	                                    timer_tick, state);
}

static void do_input_move(struct app_state *state, int x, int y)
{
	uint_fast32_t mask;

	if ((mask = game_do_move(&state->board, x, y)) != 0) {
		if (game_check_goal(&state->board) == 0) {
			int_fast32_t ms = game_finish(&state->board);
			unsigned min, sec;

			ui_timer_update(state, ms);

			sec = ms / 1000, ms %= 1000;
			min = sec / 60, sec %= 60;
			printf("You won!  Time was %u:%.2u.%.3u\n",
			       min, sec, (unsigned)ms);

			mask |= ~GOAL_MASK;
		}

		x11_redraw_game(state, mask);
	}
}

static void set_view_goal(struct app_state *state, int view_goal)
{
	state->view_goal_on_game = view_goal;
	x11_redraw_game(state, game_check_goal(&state->board));
}

static void game_input(Widget w, void *data, void *cb_data)
{
	XmDrawingAreaCallbackStruct *cbs = cb_data;
	XButtonEvent *b = &cbs->event->xbutton;
	struct app_state *state = data;
	Dimension width, height;

	switch (cbs->event->type) {
	case ButtonPress:
		switch (b->button) {
		case Button1:
			if (b->state & Button3Mask)
				break;

			XtVaGetValues(w, XmNwidth, &width,
			                 XmNheight, &height,
			                 (char *)NULL);

			do_input_move(state, b->x / (width / 5),
			                     b->y / (height / 5));
			break;
		case Button3:
			set_view_goal(state, 1);
			break;
		}
		break;
	case ButtonRelease:
		switch (b->button) {
		case Button3:
			set_view_goal(state, 0);
			break;
		}
	}
}

static struct app_state state;

static Widget get_shell(Widget w)
{
	Widget shell;

	for (shell = w; !XtIsWMShell(shell);)
		shell = XtParent(shell);

	return shell;
}

static void proc_exit(Widget w, XEvent *e, String *argv, Cardinal *argc)
{
	XtAppSetExitFlag(XtWidgetToApplicationContext(w));
}

static void proc_new_game(Widget w, XEvent *e, String *argv, Cardinal *argc)
{
	game_reset(&state.board);

	x11_redraw_goal(&state, -1);
	x11_redraw_game(&state, -1);
	x11_redraw_icon(&state, get_shell(w));

	if (!state.timer_tick) {
		XtAppContext app = XtWidgetToApplicationContext(w);
		state.timer_tick = XtAppAddTimeOut(app, TIMER_UPDATE_MS,
		                                   timer_tick, &state);
	}

	game_begin(&state.board);
}

static void proc_about(Widget w, XEvent *e, String *argv, Cardinal *argc)
{
	ui_show_about(&state, get_shell(w));
}

static const XtActionsRec menu_actions[] = {
	{ "gameNew", proc_new_game },
	{ "gameExit", proc_exit },
	{ "helpAbout", proc_about }
};

static XtAppContext app_initialize(int argc, char **argv)
{
	XtAppContext app;
	Widget shell;
	int i;

	if (argc > 0)
		progname = argv[0];

	shell = early_setup(&app, argc, argv);
	XtAppAddActions(app, (void *)menu_actions, XtNumber(menu_actions));
	ui_initialize(&state, shell);
	x11_initialize(&state, shell);

	XtAddCallback(state.game, XmNinputCallback,  game_input, &state);

	/* Begin with the game in winning state */
	game_reset(&state.board);
	for (i = 0; i < 3; i++)
		state.board.game[i] = state.board.goal[i] << GOAL_SHIFT;
	game_finish(&state.board);

	state.use_ewmh_icons = ewmh_probe_wm_icon(shell);
	XtRealizeWidget(shell);

	x11_redraw_icon(&state, shell);

	return app;
}

int main(int argc, char **argv)
{
	setlocale(LC_ALL, "");
	XtAppMainLoop(app_initialize(argc, argv));
	return 0;
}
