/*
 * X11 GUI for slide puzzle game
 * Copyright © 2022 Nick Bowler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <inttypes.h>
#include <assert.h>
#include <Xm/XmAll.h>

#include "motif.h"
#include "motifgui.h"
#include "xcounter.h"
#include "version.h"

#define SPLIT_NUMERATOR    75
#define SPLIT_DENOMINATOR 100

#define MIN(a, b) ((a) < (b) ? (a) : (b))

/* XXX generate this list? */
enum {
	widgetMainWindow,
	widgetForm,
	widgetFrame,
	widgetDrawingArea,
	widgetCascadeButton,
	widgetPushButton,
	widgetMax,

	/* Pseudo-widgets */
	widgetMenuBar = widgetMax,
	widgetMessageDialog,
	widgetScrolledText,

	widgetUnmanaged = 128
};

static WidgetClass * const widgets[widgetMax] = {
	&xmMainWindowWidgetClass,
	&xmFormWidgetClass,
	&xmFrameWidgetClass,
	&xmDrawingAreaWidgetClass,
	&xmCascadeButtonWidgetClass,
	&xmPushButtonWidgetClass
};

static const struct ui_widget {
	uint_least16_t name;
	uint_least8_t subtree;
	uint_least8_t widget_type;
} mainwin[] = { MAINWIN_INITIALIZER };

static const struct ui_menuitem {
	struct ui_widget w;
	uint_least16_t label;
} mainmenu[] = { MAINMENU_INITIALIZER };

static void
ResizeGameArea(Widget form, XEvent *e, String *args, Cardinal *num_args)
{
	Widget game = XtNameToWidget(form, &tree_strtab[gameArea]);
	Widget goal = XtNameToWidget(form, &tree_strtab[goalArea]);
	Dimension w, h, gamesz, gameborder, goalsz, goalborder;
	int x, y, gap;

	XtVaGetValues(form, XmNwidth, &w, XmNheight, &h, (char *)NULL);
	XtVaGetValues(game, XmNshadowThickness, &gameborder, (char *)NULL);
	XtVaGetValues(goal, XmNshadowThickness, &goalborder,
	                    XmNleftOffset, &gap,
	                    (char *)NULL);

	gamesz = MIN(h, w * SPLIT_NUMERATOR / SPLIT_DENOMINATOR);
	gamesz = 5*( (gamesz - 2*gameborder)/5 ) + 2*gameborder;

	goalsz = MIN(gamesz*3/5, w - gamesz - gap);
	goalsz = 3*( (goalsz - 2*goalborder)/3 ) + 2*goalborder;

	x = (w - gamesz - goalsz - gap) / 2;
	if (x < 2) x = 0;

	y = (h - gamesz) / 2;
	if (y < 3) y = 0;

	XtVaSetValues(game, XmNleftOffset, x, XmNtopOffset, y, (char *)NULL);
	XtVaSetValues(game, XmNwidth, gamesz, XmNheight, gamesz, (char *)NULL);
	XtVaSetValues(goal, XmNwidth, goalsz, XmNheight, goalsz, (char *)NULL);
}

static char *timer_text(int_fast32_t ms, char *buf)
{
	unsigned min, sec;

	sec = ms / 1000, ms %= 1000;
	min = sec / 60, sec %= 60;
	sprintf(buf, "Time: %u:%.2u.%.3u", min, sec, (unsigned)ms);

	return buf;
}

void ui_timer_update(struct app_state *state, int_fast32_t ms)
{
	char buf[100];

	if (ms < 0) {
		xcounter_simple_update(state->timer, "\n");
		return;
	}

	xcounter_simple_update(state->timer, timer_text(ms, buf));
}

static void configure_mainwin(struct app_state *state, Widget form)
{
	Widget gamearea = XtNameToWidget(form, &tree_strtab[gameArea]);
	Widget goalarea = XtNameToWidget(form, &tree_strtab[goalArea]);
	XtActionsRec resize_rec;
	char xc_template[100];

	assert(gamearea && goalarea);
	XtVaSetValues(form, XmNfractionBase, SPLIT_DENOMINATOR, (char *)NULL);

	state->game = XtNameToWidget(gamearea, &tree_strtab[gameCanvas]);
	XtVaSetValues(gamearea, XmNleftAttachment, XmATTACH_FORM,
	                        XmNtopAttachment, XmATTACH_FORM,
	                        (char *)NULL);

	state->goal = XtNameToWidget(goalarea, &tree_strtab[goalCanvas]);
	XtVaSetValues(goalarea, XmNleftAttachment, XmATTACH_WIDGET,
	                        XmNleftWidget, gamearea,
	                        XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
	                        XmNtopWidget, gamearea,
	                        (char *)NULL);

	state->timer = XtNameToWidget(form, &tree_strtab[timeDisplay]);
	XtVaSetValues(state->timer, XmNleftAttachment, XmATTACH_WIDGET,
	                            XmNleftWidget, gamearea,
	                            XmNtopAttachment, XmATTACH_WIDGET,
	                            XmNtopWidget, goalarea,
	                            XmNrightAttachment, XmATTACH_FORM,
	                            (char *)NULL);

	xcounter_simple_setup(state->timer, timer_text(20000, xc_template));
	ui_timer_update(state, -1);

	resize_rec.string = "ResizeGameArea";
	resize_rec.proc = ResizeGameArea;
	XtAppAddActions(XtWidgetToApplicationContext(form), &resize_rec, 1);
	XtOverrideTranslations(form, XtParseTranslationTable(
		"<Configure>: ResizeGameArea()\n"
		"<Map>: ResizeGameArea()\n"
	));

	/*
	 * Performing the initial update of the layout seems to avoid
	 * some weird problems on Motif 2.1
	 */
	ResizeGameArea(form, 0, 0, 0);
}

static Widget create_widget(const struct ui_widget *item, Widget parent,
                            ArgList args, Cardinal num_args)
{
	String name = (void *)&tree_strtab[item->name];
	unsigned type = item->widget_type & widgetUnmanaged-1;

	switch (type) {
	case widgetMenuBar:
		return XmCreateMenuBar(parent, name, args, num_args);
	case widgetMessageDialog:
		return XmCreateMessageDialog(parent, name, args, num_args);
	case widgetScrolledText:
		return XmCreateScrolledText(parent, name, args, num_args);
	}

	assert(type < widgetMax);
	return XtCreateWidget(name, *widgets[type], parent, args, num_args);
}

static void
construct_widgets(const struct ui_widget *root, Widget parent, unsigned i)
{
	const struct ui_widget *item;

	for (item = &root[i]; item->name; item++) {
		Widget w = create_widget(item, parent, NULL, 0);

		if (item->subtree)
			construct_widgets(root, w, item->subtree);

		if (!(item->widget_type & widgetUnmanaged))
			XtManageChild(w);
	}
}

static void menu_cb(Widget w, void *data, void *cb_data)
{
	XmRowColumnCallbackStruct *cbs = cb_data;
	XtCallActionProc(cbs->widget, XtName(cbs->widget), cbs->event, NULL, 0);
}

static Widget create_pulldown(Widget parent)
{
	Widget w;

	w = XmCreatePulldownMenu(parent, XtName(parent), NULL, 0);
	XtVaSetValues(parent, XmNsubMenuId, w, (char *)NULL);
	XtAddCallback(w, XmNentryCallback, menu_cb, NULL);

	return w;
}

static void
construct_menu(const struct ui_menuitem *root, Widget parent, unsigned i)
{
	const struct ui_menuitem *item;

	for (item = &root[i]; item->w.name; item++) {
		const char *label = &strtab[item->label];
		unsigned n = 0;
		Arg args[2];
		XmString s;
		Widget w;

		if (XtClass(parent) == *widgets[widgetCascadeButton])
			parent = create_pulldown(parent);

		if (label[0] && label[1] == '|') {
			XtSetArg(args[n], XmNmnemonic, label[0]); n++;
			label += 2;
		}

		s = XmStringCreateLocalized((void *)label);
		XtSetArg(args[n], XmNlabelString, s); n++;

		w = create_widget(&item->w, parent, args, n);

		XmStringFree(s);

		if (item->w.subtree)
			construct_menu(root, w, item->w.subtree);

		XtManageChild(w);
	}
}

/* Figure out which tiles intersect a rectangle. */
static uint_fast32_t
expose_mask(int rect_x, int rect_y, int rect_w, int rect_h,
                                    int tile_w, int tile_h)
{
	return board_right(rect_x / tile_w)
	     & board_below(rect_y / tile_h)
	     & board_above((rect_y + rect_h - 1) / tile_h)
	     & board_left((rect_x + rect_w - 1) / tile_w);
}

/*
 * Calculate which tiles need to be redrawn after resizing.  This is the
 * tiles which (at the new size) are fully contained in the previous area.
 *
 * When shrinking, this is all tiles, but when enlarging, the generated
 * expose events will trigger drawing of the new area.
 */
static uint_fast32_t resize_helper(Widget w, Dimension *oldsz, int gridsz)
{
	Dimension new_w, new_h, common_w, common_h;

	if (!XtIsRealized(w))
		return 0;

	XtVaGetValues(w, XmNwidth, &new_w, XmNheight, &new_h, (char *)NULL);
	common_w = MIN(new_w, oldsz[0]);
	common_h = MIN(new_h, oldsz[1]);
	oldsz[0] = new_w;
	oldsz[1] = new_h;
	new_w /= gridsz;
	new_h /= gridsz;

	/* Round down to multiple of tile dimensions */
	common_w = common_w / new_w * new_w;
	common_h = common_h / new_h * new_h;

	if (new_w > 0 && new_h > 0)
		return expose_mask(0, 0, common_w, common_h, new_w, new_h);
	return 0;
}

static void game_resize(Widget w, void *data, void *cb_data)
{
	struct app_state *state = data;
	uint_fast32_t mask;

	mask = resize_helper(w, state->game_sz, 5);
	if (mask)
		x11_redraw_game(data, mask);
}

static void goal_resize(Widget w, void *data, void *cb_data)
{
	struct app_state *state = data;
	uint_fast32_t mask;

	mask = resize_helper(w, state->goal_sz, 3);
	if (mask)
		x11_redraw_goal(data, mask);
}

static void game_expose(Widget w, void *data, void *cb_data)
{
	XmDrawingAreaCallbackStruct *cbs = cb_data;
	XExposeEvent *e = &cbs->event->xexpose;
	Dimension width, height;
	uint_fast32_t mask;

	XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, (char *)NULL);
	if (!(width /= 5) || !(height /= 5))
		return;

	mask = expose_mask(e->x, e->y, e->width, e->height, width, height);
	x11_redraw_game(data, mask);
}

static void goal_expose(Widget w, void *data, void *cb_data)
{
	XmDrawingAreaCallbackStruct *cbs = cb_data;
	XExposeEvent *e = &cbs->event->xexpose;
	Dimension width, height;
	uint_fast32_t mask;

	XtVaGetValues(w, XmNwidth, &width, XmNheight, &height, (char *)NULL);
	if (!(width /= 3) || !(height /= 3))
		return;

	mask = expose_mask(e->x, e->y, e->width, e->height, width, height);
	x11_redraw_goal(data, mask);
}

void ui_initialize(struct app_state *state, Widget shell)
{
	Widget menubar, help;

	construct_widgets(mainwin, shell, 0);

	menubar = XtNameToWidget(shell, "*menuBar");
	construct_menu(mainmenu, menubar, 0);

	help = XtNameToWidget(menubar, "helpMenu");
	XtVaSetValues(menubar, XmNmenuHelpWidget, help, (char *)NULL);

	configure_mainwin(state, XtNameToWidget(shell, "*game"));

	XtVaGetValues(state->game, XmNwidth, &state->game_sz[0],
	                           XmNheight, &state->game_sz[1],
	                           (char *)NULL);
	XtAddCallback(state->game, XmNresizeCallback, game_resize, state);
	XtAddCallback(state->game, XmNexposeCallback, game_expose, state);

	XtVaGetValues(state->game, XmNwidth, &state->goal_sz[0],
	                           XmNheight, &state->goal_sz[1],
	                           (char *)NULL);
	XtAddCallback(state->goal, XmNresizeCallback, goal_resize, state);
	XtAddCallback(state->goal, XmNexposeCallback, goal_expose, state);
}

static void dialog_close(Widget w, void *data, void *cb_data)
{
	XtDestroyWidget(w);
}

/*
 * Expands to an XtVaSetValues argument list to set the given resource with
 * a typed string value, which is internally converted to the appropriate
 * resource type.
 *
 * Note that str is expanded twice and thus should not have side effects.
 */
#define STRING_ARG(resource, str) \
	XtVaTypedArg, (resource), XtRString, (str), strlen((str))+1

void ui_show_about(struct app_state *state, Widget shell)
{
	static const struct ui_widget dialog[] = { ABOUTDIALOG_INITIALIZER };
	Widget w, l;
	char *msg;

	construct_widgets(dialog, shell, 0);

	w = XtNameToWidget(shell, "*aboutDialog");
	XtUnmanageChild(XmMessageBoxGetChild(w, XmDIALOG_CANCEL_BUTTON));

	XtAddCallback(w, XmNunmapCallback, dialog_close, NULL);

	XtVaSetValues(w, STRING_ARG(XmNokLabelString, "Close"), (char *)NULL);

	msg = version_format_head("rrace-motif");
	l = XmMessageBoxGetChild(w, XmDIALOG_MESSAGE_LABEL);
	XtVaSetValues(l, XmNlabelType, XmPIXMAP_AND_STRING,
	                 XmNlabelPixmap, state->icon_pixmap,
	                 STRING_ARG(XmNlabelString, msg),
	                 (char *)NULL);
	free(msg);

	l = XtNameToWidget(w, "*licenseBlurb");
	XmTextSetString(l,
"This program is free software: you can redistribute it and/or modify\n"
"it under the terms of the GNU General Public License as published by\n"
"the Free Software Foundation, either version 3 of the License, or\n"
"(at your option) any later version.\n\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program.  If not, see <https://www.gnu.org/licenses/>.");

	XtVaSetValues(l, XmNeditMode, XmMULTI_LINE_EDIT,
	                 XmNeditable, FALSE,
	                 XmNresizeWidth, TRUE,
			 XmNrows, 5,
	                 (char *)NULL);

	XtManageChild(w);
}
